/*
 * pu-buff.c - buffer management
 *
 * Copyright (C) 2011..2012  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <pu-buff.h>
#include <pu-lib.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#if 0
#define DEBUG	1
#endif

#ifdef DEBUG
#define debug(fmt...) fprintf(stderr, fmt);
#else
#define debug(fmt...) do { } while (0)
#endif

struct pu_buff *pu_alloc_skb(unsigned int size, pu_gfp_t priority)
{
	struct pu_buff *skb = xmalloc(sizeof(*skb));

	skb->head = xmalloc(size);
	skb->data = skb->head;
	skb->tail = 0;
	skb->end = size;
	skb->len = 0;
	skb->data_len = 0;
	skb->users = 1;
	return skb;
}

struct pu_buff *pu_skb_get(struct pu_buff *skb)
{
	skb->users++;
	return skb;
}

void pu_dump_skb(struct pu_buff *skb)
{
#if DEBUG
	fprintf(stderr, "skb %p: len: %d data: %d tail: %d, "
			"%02x %02x %02x %02x\n",
			skb, skb->len,
			skb->data - skb->head, skb->tail,
			skb->data[0], skb->data[1], skb->data[2], skb->data[3]);
#endif
}


/**
 *	skb_headroom - bytes at buffer head
 *	@skb: buffer to check
 *
 *	Return the number of bytes of free space at the head of an &sk_buff.
 */
unsigned int pu_skb_headroom(const struct pu_buff *skb)
{
	return skb->data - skb->head;
}

/**
 *	skb_tailroom - bytes at buffer end
 *	@skb: buffer to check
 *
 *	Return the number of bytes of free space at the tail of an sk_buff
 */
int pu_skb_tailroom(const struct pu_buff *skb)
{
	return skb->end - skb->tail;
}

void pu_skb_reserve(struct pu_buff *skb, int len)
{
	skb->data += len;
	skb->tail += len;
}

unsigned char *pu_skb_tail_pointer(const struct pu_buff *skb)
{
	return skb->head + skb->tail;
}

void pu_skb_reset_tail_pointer(struct pu_buff *skb)
{
	skb->tail = skb->data - skb->head;
}

void pu_skb_set_tail_pointer(struct pu_buff *skb, const int offset)
{
	pu_skb_reset_tail_pointer(skb);
	skb->tail += offset;
}

unsigned char *skb_tail_pointer(const struct pu_buff *skb)
{
	return skb->head + skb->tail;
}

static void pu_skb_under_panic(struct pu_buff *skb, int sz)
{
	die("skb underflow %d, end: %d", sz, skb->end);
}

static void pu_skb_over_panic(struct pu_buff *skb, int sz)
{
	die("skb overflow %d, end: %d", sz, skb->end);
}

static void __pu_skb_trim(struct pu_buff *skb, unsigned int len)
{
	skb->len = len;
	pu_skb_set_tail_pointer(skb, len);
}

void pu_skb_trim(struct pu_buff *skb, unsigned int len)
{
	if (skb->len > len)
		__pu_skb_trim(skb, len);
}

unsigned char *pu_skb_put(struct pu_buff *skb, unsigned int len)
{
	unsigned char *tmp = skb_tail_pointer(skb);

	skb->tail += len;
	skb->len += len;
	if (unlikely(skb->tail > skb->end))
		pu_skb_over_panic(skb, len);
	return tmp;
}

unsigned char *pu_skb_push(struct pu_buff *skb, unsigned int len)
{
	skb->data -= len;
	skb->len += len;
	if (unlikely(skb->data < skb->head))
		pu_skb_under_panic(skb, len);
	return skb->data;
}

unsigned char *pu_skb_pull(struct pu_buff *skb, unsigned int len)
{
	skb->len -= len;
	if (skb->len < skb->data_len)
		die("pu_skb_pull: underflow %d", len);
	return skb->data += len;
}

struct pu_buff *pu_skb_copy(struct pu_buff *src)
{
	struct pu_buff *skb;

	skb = pu_alloc_skb(src->end, GFP_ATOMIC);
	if (skb != NULL) {
		skb->tail = src->tail;
		skb->end = src->end;
		skb->len = src->len;
		skb->data_len = src->data_len;
		memcpy(skb->head, src->head, src->end);
		skb->data = skb->head + (src->data - src->head);
	}
	return skb;
}

struct pu_buff *skb_copy_expand(struct pu_buff *skb, int newheadroom,
		int newtailroom, pu_gfp_t priority)
{
	struct pu_buff *n = pu_alloc_skb(newheadroom + skb->len + newtailroom,
			priority);
	int oldheadroom = pu_skb_headroom(skb);
	int head_copy_len, head_copy_off;

	if (!n)
		return NULL;

	pu_skb_reserve(n, newheadroom);

	/* Set the tail pointer and length */
	pu_skb_put(n, skb->len);

	head_copy_len = oldheadroom;
	head_copy_off = 0;
	if (newheadroom <= head_copy_len)
		head_copy_len = newheadroom;
	else
		head_copy_off = newheadroom - head_copy_len;

	memcpy(n->data + head_copy_off, skb->data - head_copy_len,
			skb->len + head_copy_len);
	return n;
}

void pu_consume_skb(struct pu_buff *skb)
{
	skb->users--;
	if (!skb->users) {
		debug("skb: freeing %p head=%p\n", skb, skb->head);
		xfree(skb->head);
		xfree(skb);
	}
}
