/*
 * manhattan - pu-get.c
 *
 * Download data chunk using PU protocol
 *
 * Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pu-lib.h>
#include <pu-buff.h>
#include <pu-proto.h>
#include <pu-proto-lib.h>
#include <list.h>

#ifndef PU_MTU
#define PU_MTU		1048576
#endif

static const char *server_host = "localhost";
static const char *server_port = "3000";

static int download_all;
static int follow;
static int use_fsync;
static int download_igroup;

static int written_samples;
static int total_samples;

struct sample {
	unsigned int id;
	struct list_head list;
};
static LIST_HEAD(to_download);

static const char copyright_string[] =
"pu-get " PACKAGE_VERSION ".\n"
"Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>.\n";

static void show_copyright(void)
{
	puts(copyright_string);
}
	
static const char usage[] =
"usage: %s [options] [--] <id...>\n"
"Options:\n"
"      --fsync                use fsync()\n"
"  -V, --version              show version\n"
"  -a, --all                  download all available data\n"
"  -f, --follow               download data as the available data grows\n"
"      --igroup               write only important groups\n"
"  -h, --help                 show this help\n"
"  -p, --port=PORT            use specified port (3000)\n"
"  -s, --host=HOST            use specified target host (localhost)\n"
"\n"
"Report bugs to krzysiek@podlesie.net.\n";

static void show_usage(char *name)
{
	show_copyright();

	printf(usage, (name == NULL) ? "pu-get" : name);
	exit(EXIT_SUCCESS);
}

static const char license[] =
"This program comes with ABSOLUTELY NO WARRANTY; see COPYING for details.\n"
"This is free software, and you are welcome to redistribute it\n"
"under certain conditions; see COPYING for details.\n";

static void show_version(void)
{
	show_copyright();

	puts(license);
	exit(EXIT_SUCCESS);
}

static void queue_get(const char *x)
{
	struct sample *b;
	char *end;
	long id;

	if (*x == '\0') {
		fprintf(stderr, "warning: skipping empty id\n");
		return;
	}

	id = strtol(x, &end, 0);
	if (*end != '\0') {
		fprintf(stderr, "error: invalid id '%s'\n", x);
		exit(EXIT_FAILURE);
	}

	if (id < 0) {
		fprintf(stderr, "error: id '%s' must be non-negative\n", x);
		exit(EXIT_FAILURE);
	}

	b = xmalloc(sizeof(*b));
	b->id = id;
	list_add(&b->list, &to_download);
}

static void parse_command_line(int argc, char **argv)
{
	static struct option long_options[] = {
		{"all", 0, NULL, 'a'},
		{"follow", 0, NULL, 'f'},
		{"fsync", 0, NULL, 1},
		{"igroup", 0, NULL, 2},
		{"help", 0, NULL, 'h'},
		{"host", 1, NULL, 's'},
		{"port", 1, NULL, 'p'},
		{"version", 0, NULL, 'V'},
		{NULL, 0, NULL, 0},
	};
	int option_index = 0;
	int c;

	while (1) {
		c = getopt_long(argc, argv, "afhp:s:V",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 1:
			use_fsync = 1;
			break;
		case 2:
			download_igroup = 1;
			break;
		case 'a':
			download_all = 1;
			break;
		case 'f':
			follow = 1;
			break;
		case 'h':
			show_usage(argv[0]);
			BUG();
		case 'p':
			server_port = strdup(optarg);
			break;
		case 's':
			server_host = strdup(optarg);
			break;
		case 'V':
			show_version();
			BUG();
		default:
			fprintf(stderr, "Unknown option: %d.\n", c);
			exit(EXIT_FAILURE);
		}
	}

	while (optind < argc)
		queue_get(argv[optind++]);
}

static int do_connect(const char *host, const char *port)
{
	struct addrinfo hints, *ai, *aitop;
	int ret;
	int fd;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = 0;
	ret = getaddrinfo(host, port, &hints, &aitop);
	if (ret < 0) {
		fprintf(stderr, "getaddrinfo(%s, %s): %s\n", host, port,
				strerror(errno));
		exit(EXIT_FAILURE);
	}

	for (ai = aitop; ai; ai = ai->ai_next) {
		fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
		if (fd == -1) {
			fprintf(stderr, "warning: socket: %s\n",
					strerror(errno));
			continue;
		}

		ret = connect(fd, ai->ai_addr, ai->ai_addrlen);
		if (ret < 0) {
			fprintf(stderr, "warning: connect: %s\n",
					strerror(errno));
			close(fd);
			continue;
		}
		freeaddrinfo(aitop);
		return fd;
	}
	fprintf(stderr, "error: can't connect to %s:%s\n", host, port);
	exit(EXIT_FAILURE);
}

int pu_proto_xmit(void *priv, struct pu_buff *skb)
{
	int fd = *((int *) priv);

	write_or_die(fd, skb->data, skb->len);
	pu_consume_skb(skb);
	return 0;
}

static struct pu_buff *pu_proto_recv(void *priv)
{
	struct pu_buff *skb;
	struct pu_header *hdr;
	int fd = *((int *) priv);
	unsigned long len;
	ssize_t rd;

	skb = pu_alloc_skb(512, GFP_KERNEL);
	rd = read_in_full(fd, pu_skb_put(skb, sizeof(*hdr)), sizeof(*hdr));
	if (rd != sizeof(*hdr))
		die_errno("cannot read header");

	hdr = (void *) skb->data;
	if (hdr->magic != PU_MAGIC)
		die("protocol error: invalid magic %02x", hdr->magic);

	len = be32_to_cpu(hdr->data_bytes);
	if ((len > 0) && !(len & PU_DATA_ERROR)) {
		unsigned int align = pu_align_data(len);
		struct pu_buff *skb2;

		if (align >= PU_MTU)
			die("message to large: %d bytes", len);

		skb2 = skb_copy_expand(skb, 0, align, GFP_KERNEL);
		pu_consume_skb(skb);
		skb = skb2;

		rd = read_in_full(fd, pu_skb_put(skb, align), align);
		if (rd != align)
			die_errno("cannot read payload");

		pu_skb_trim(skb, sizeof(*hdr) + len);
	}
	return skb;
}

static void *pu_proto_verify(struct pu_buff *skb, unsigned int proto, unsigned int id)
{
	struct pu_header *hdr = (void *) skb->data;
	unsigned int pkt_id;
	unsigned int len;

	if (hdr->proto != proto) {
		die("protocol error: invalid subprotocol: %d, expected %d",
				hdr->proto, proto);
	}

	pkt_id = be16_to_cpu(hdr->id);
	if (pkt_id != id) {
		die("protocol error: invalid id: %d, expected %d",
				pkt_id, id);
	}

	len = be32_to_cpu(hdr->data_bytes);
	if (len & PU_DATA_ERROR) {
		fprintf(stderr, "protocol warning: id: %d, remote error: %d\n",
				pkt_id, len & ~PU_DATA_ERROR);
		return NULL;
	}

	return pu_skb_pull(skb, sizeof(*hdr));
}

static void show_stat(int fd)
{
	struct pu_buff *skb;
	struct pu_stat_resp *resp;
	unsigned int current;
	unsigned int first;
	unsigned int igroup;
	unsigned int generation;
	unsigned int sample_size;
	unsigned int samples_per_group;

	skb = pu_proto_alloc_skb(0);
	pu_proto_xmit_new(&fd, skb, PU_PROTO_CMD, PU_CMD_STAT);

	skb = pu_proto_recv(&fd);
	resp = pu_proto_verify(skb, PU_PROTO_CMD, PU_CMD_STAT);
	if (!resp)
		die("stat error");
	pu_skb_pull(skb, sizeof(*resp));

	current = be32_to_cpu(resp->current);
	first = be32_to_cpu(resp->first);
	igroup = be32_to_cpu(resp->igroup);
	generation = be32_to_cpu(resp->generation);
	sample_size = be32_to_cpu(resp->sample_size);
	samples_per_group = be32_to_cpu(resp->samples_per_group);

	printf("range: %u to %u\n", first, current);
	printf("first: %u\n", first);
	printf("current: %u\n", current);
	printf("igroup: %u\n", igroup);
	printf("generation: %u\n", generation);
	printf("sample_size: %u\n", sample_size);
	printf("samples_per_group: %u\n", samples_per_group);

	pu_consume_skb(skb);
}

static void write_sample(unsigned int id, struct pu_buff *skb)
{
	char fname[PATH_MAX];
	ssize_t wr;
	int ret;
	int fd;

	snprintf(fname, PATH_MAX, "%08x.pu", id);
	fd = open(fname, O_WRONLY | O_CREAT | O_TRUNC, 0666);
	if (fd == -1) {
		fprintf(stderr, "warning: cannot create '%s'\n", fname);
		return;
	}

	wr = write_in_full(fd, skb->data, skb->len);
	if (wr != skb->len) {
		if (wr < 0) {
			fprintf(stderr, "%s: write error: %s\n",
					fname, strerror(errno));
		} else {
			fprintf(stderr, "%s: written %d of %d bytes\n",
					fname, wr, skb->len);
		}
		return;
	}

	if (use_fsync) {
		ret = fsync(fd);
		if (ret) {
			fprintf(stderr, "%s: fsync error: %s\n",
					fname, strerror(errno));
			return;
		}
	}

	ret = close(fd);
	if (ret) {
		fprintf(stderr, "%s: close error: %s\n",
				fname, strerror(errno));
		return;
	}

	written_samples++;
}

static void download(int fd, unsigned int id)
{
	struct pu_buff *skb;
	struct pu_read *rd;

	skb = pu_proto_alloc_skb(sizeof(*rd));
	rd = (void *) pu_skb_put(skb, sizeof(*rd));
	rd->sample = cpu_to_be32(id);
	pu_proto_xmit_new(&fd, skb, PU_PROTO_CMD, PU_CMD_READ);

	skb = pu_proto_recv(&fd);
	if (pu_proto_verify(skb, PU_PROTO_CMD, PU_CMD_READ))
		write_sample(id, skb);

	pu_consume_skb(skb);
}

static const char pack_hdr_fmt[] =
"pu-pack\n"
"version: 0\n"
"samples: %d %d\n";

static const char pack_sample_fmt[] =
"\n%lx sample %x";

static int download_pack(int fd, unsigned int from, unsigned int to)
{
	struct pu_buff *skb;
	struct pu_read *rd;
	unsigned int n;
	unsigned int i;
	char buf[4096];
	char fname[PATH_MAX];
	int pack_fd;
	ssize_t wr;
	int ret;

	snprintf(fname, PATH_MAX, "%08x.pu-pack", from);
	pack_fd = open(fname, O_WRONLY | O_CREAT | O_EXCL | O_TRUNC, 0666);
	if (pack_fd == -1) {
		if (!follow) {
			fprintf(stderr, "warning: cannot create '%s': %s\n",
					fname, strerror(errno));
		} else {
			if (errno != EEXIST) {
				die("cannot create '%s': %s\n", fname,
						strerror(errno));
			}
		}
		return -errno;
	}

	n = snprintf(buf, sizeof(buf), pack_hdr_fmt, from, to) + 1;
	wr = write_in_full(pack_fd, buf, n);
	if (wr != n) {
		ret = errno;
		if (wr < 0) {
			fprintf(stderr, "%s: write error: %s\n",
					fname, strerror(errno));
		} else {
			fprintf(stderr, "%s: written %d of %d bytes\n",
					fname, wr, n);
		}
		goto err_unlink;
	}

	for (i = from; i < to; i++) {
		skb = pu_proto_alloc_skb(sizeof(*rd));
		rd = (void *) pu_skb_put(skb, sizeof(*rd));
		rd->sample = cpu_to_be32(i);
		pu_proto_xmit_new(&fd, skb, PU_PROTO_CMD, PU_CMD_READ);

		skb = pu_proto_recv(&fd);
		if (!pu_proto_verify(skb, PU_PROTO_CMD, PU_CMD_READ)) {
			ret = -EPROTO;
			fprintf(stderr, "%s: PU_CMD_READ error\n", fname);
			goto err_consume_skb;
		}

		n = snprintf(buf, sizeof(buf), pack_sample_fmt,
				(unsigned long) skb->len, i) + 1;
		wr = write_in_full(pack_fd, buf, n);
		if (wr != n) {
			ret = errno;
			if (wr < 0) {
				fprintf(stderr, "%s: write error: %s\n",
						fname, strerror(errno));
			} else {
				fprintf(stderr, "%s: written %d of %d bytes\n",
						fname, wr, n);
			}
			goto err_consume_skb;
		}

		wr = write_in_full(pack_fd, skb->data, skb->len);
		if (wr != skb->len) {
			if (wr < 0) {
				ret = errno;
				fprintf(stderr, "%s: write error: %s\n",
						fname, strerror(errno));
			} else {
				ret = -EINVAL;
				fprintf(stderr, "%s: written %d of %d bytes\n",
						fname, wr, skb->len);
			}
			goto err_consume_skb;
		}

		pu_consume_skb(skb);
	}

	if (use_fsync) {
		ret = fsync(pack_fd);
		if (ret) {
			fprintf(stderr, "%s: fsync error: %s\n",
					fname, strerror(errno));
		}
	}

	ret = close(pack_fd);
	if (ret) {
		fprintf(stderr, "%s: close error: %s\n",
				fname, strerror(errno));
	}

	return ret;

err_consume_skb:
	pu_consume_skb(skb);
err_unlink:
	unlink(fname);
	return ret;
}

static void download_list(int fd)
{
	struct sample *s;

	while (!list_empty(&to_download)) {
		s = list_entry(to_download.next, struct sample, list);
		total_samples++;

		download(fd, s->id);
		list_del(&s->list);
		xfree(s);
	}

	fprintf(stderr, "%d/%d samples sucessfuly saved\n",
			written_samples, total_samples);
}

#ifndef warning
#define warning(fmt...) fprintf(stderr, "warning: " fmt);
#endif

static void do_download_all(int fd)
{
	struct pu_buff *skb;
	struct pu_stat_resp *resp;
	unsigned int i = 0;
	unsigned int current;
	unsigned int first;

	do {
		skb = pu_proto_alloc_skb(0);
		pu_proto_xmit_new(&fd, skb, PU_PROTO_CMD, PU_CMD_STAT);

		skb = pu_proto_recv(&fd);
		resp = pu_proto_verify(skb, PU_PROTO_CMD, PU_CMD_STAT);
		if (!resp)
			die("stat error");
		pu_skb_pull(skb, sizeof(*resp));

		current = be32_to_cpu(resp->current);
		first = be32_to_cpu(resp->first);
		pu_consume_skb(skb);

		if (first > current)
			die("first > current");

		if (i > current) {
			warning("sequence number rewinded %d -> %d\n",
					i, current);
			i = 0;
		}
		if (i < first)
			i = first;

		for (; i < current; i++) {
			download(fd, i);
			total_samples++;
		}
		if (follow)
			sleep(1);
	} while (follow);

	fprintf(stderr, "%d/%d samples sucessfuly saved\n",
			written_samples, total_samples);
}

static void do_igroup_download(int fd)
{
	struct pu_buff *skb;
	struct pu_stat_resp *resp;
	unsigned int i;
	unsigned int current;
	unsigned int first;
	unsigned int igroup;
	unsigned int samples_per_group;
	skb = pu_proto_alloc_skb(0);
	pu_proto_xmit_new(&fd, skb, PU_PROTO_CMD, PU_CMD_STAT);

	skb = pu_proto_recv(&fd);
	resp = pu_proto_verify(skb, PU_PROTO_CMD, PU_CMD_STAT);
	if (!resp)
		die("stat error");
	pu_skb_pull(skb, sizeof(*resp));

	current = be32_to_cpu(resp->current);
	first = be32_to_cpu(resp->first);
	igroup = be32_to_cpu(resp->igroup);
	samples_per_group = be32_to_cpu(resp->samples_per_group);
	pu_consume_skb(skb);

	if (first > current)
		die("first > current");
	if (igroup > current)
		die("igroup > current");

	if (!igroup) {
		info("igroup not present");
		return;
	}

	if (igroup < samples_per_group)
		die("igroup < samples_per_group");

	i = igroup - samples_per_group;
	if (i < first) {
		warn("igroup %d, no longer available [%d; %d)",	igroup,
				first, current);
		return;
	}

	download_pack(fd, i, i + samples_per_group);
}

static void igroup_download(int fd)
{
	do {
		do_igroup_download(fd);
		if (follow)
			sleep(1);
	} while (follow);
}

static void pu_close(int fd)
{
	int warned = 0;
	ssize_t rd;
	char c;

	shutdown(fd, SHUT_WR);
	do {
		rd = xread(fd, &c, sizeof(c));
		if (rd < 0)
			die_errno("read error");
		if (rd && !warned) {
			fprintf(stderr, "warning: received garbage\n");
			warned = 1;
		}
	} while (rd);
	close(fd);
}

int main(int argc, char **argv)
{
	int fd;

	parse_command_line(argc, argv);

	fd = do_connect(server_host, server_port);
	if (download_igroup) {
		igroup_download(fd);
	} else if (!list_empty(&to_download))
		download_list(fd);
	else if (download_all) {
		do_download_all(fd);
	} else {
		show_stat(fd);
	}
	pu_close(fd);
	return 0;
}
