/*
 * manhattan - pu-sim.c
 *
 * DSP simulator
 *
 * Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <sys/mman.h>
#include <pu-lib.h>
#include <pu-buff.h>
#include <asm_io.h>
#include <pu-shm.h>

static const char *dsp_device = "/dev/zero";
static unsigned long long dsp_offset;
static unsigned long long dsp_size = 24 * 1048576;
static unsigned long long dsp_sample = 1024 * 10 * 4;
static unsigned int samples_per_group = 480;

static u8 *shm;
static unsigned int sample_size;
static unsigned int sample_align;
static unsigned int ring_entries;

static unsigned long blocks_from;
static unsigned long blocks_to;

static unsigned long interval = 1000000;

static const char copyright_string[] =
"pu-sim " PACKAGE_VERSION ".\n"
"Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>.\n";

static void show_copyright(void)
{
	puts(copyright_string);
}

static const char usage[] =
"usage: %s [options] [FROM [TO]]\n"
"\n"
"Simulate shared memory writes with block numbers from FROM to TO.\n"
"\n"
"Options:\n"
"  -V, --version              show version\n"
"  -f, --file=FILE            shared memory file\n"
"  -i, --interval=INTEVAL     use sample generation interval [ms]\n"
"  -h, --help                 show this help\n"
"  -n, --sample=SAMPLE        sample size\n"
"  -o, --offset=OFFSET        offset in shared memory\n"
"  -s, --shmsize=SIZE         shared memory size\n"
"\n"
"Report bugs to krzysiek@podlesie.net/\n";

static void show_usage(char *name)
{
	show_copyright();

	printf(usage, (name == NULL) ? "pu-sim" : name);
	exit(EXIT_SUCCESS);
}

static const char license[] =
"This program comes with ABSOLUTELY NO WARRANTY; see COPYING for details.\n"
"This is free software, and you are welcome to redistribute it\n"
"under certain conditions; see COPYING for details.\n";

static void show_version(void)
{
	show_copyright();

	puts(license);
	exit(EXIT_SUCCESS);
}

static void parse_command_line(int argc, char **args)
{
	static struct option long_options[] = {
		{"file", 1, NULL, 'f'},
		{"interval", 1, NULL, 'i'},
		{"help", 0, NULL, 'h'},
		{"memsize", 1, NULL, 's'},
		{"offset", 1, NULL, 'o'},
		{"sample", 1, NULL, 'n'},
		{"version", 0, NULL, 'V'},
		{NULL, 0, NULL, 0},
	};
	int option_index = 0;
	int c;

	while (1) {
		c = getopt_long(argc, args, "f:i:ho:n:s:V",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 'f':
			dsp_device = strdup(optarg);
			break;
		case 'i':
			interval = 1000 * atoll(optarg);
			break;
		case 'h':
			show_usage(args[0]);
			BUG();
		case 'n':
			dsp_sample = atoll(optarg);
			break;
		case 'o':
			dsp_offset = atoll(optarg);
			break;
		case 's':
			dsp_size = atoll(optarg);
			break;
		case 'V':
			show_version();
			BUG();
		default:
			fprintf(stderr, "Unknown option.\n");
			exit(EXIT_FAILURE);
		}
	}

	if (optind >= argc)
		return;
	blocks_from = atoll(args[optind++]);
	if (optind >= argc)
		return;
	blocks_to = atoll(args[optind++]);
}

static int dspshm_init(const char *device, unsigned long long offset,
		unsigned long size, unsigned int sample)
{
	unsigned long page_size = sysconf(_SC_PAGE_SIZE);
	void *p;
	int ret;
	int fd;

	ret = -EINVAL;
	if (size & (page_size - 1)) {
		fprintf(stderr, "error: size must be page aligned\n");
		goto out;
	}

	ret = -EINVAL;
	if (offset & (page_size - 1)) {
		fprintf(stderr, "error: offset must be page aligned\n");
		goto out;
	}

	/*
	 * read-only access is currently sufficient, we just want to
	 * watch DSP sequence numbers and receive data blocks
	 */
	fd = open(device, O_RDWR);
	if (fd == -1) {
		perror("cannot open DSP device");
		ret = -errno;
		goto out;
	}

	p = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, offset);
	if (p == MAP_FAILED) {
		perror("cannot map DSP shared memory");
		ret = -errno;
		goto out_close;
	}
	sample_size = sample;
	sample_align = (sample + DSP_ALIGNMENT - 1) & ~(DSP_ALIGNMENT - 1);
	ring_entries = (size - DSP_RING_BASE) / sample_align;
	fprintf(stderr, "dspcomm: shm: %ld KiB, sample %dB, %d entries\n",
			size >> 10, sample_size, ring_entries);
	ret = -EINVAL;
	if (ring_entries < 2) {
		fprintf(stderr, "too small ring size, need >= 2 entries\n");
		munmap(p, size);
		goto out_close;
	}
	shm = p;
	ret = 0;
out_close:
	close(fd);
out:
	return ret;
}

static u8 *dsp_get_buffer(unsigned int seq)
{
	unsigned long offset;

	seq %= ring_entries;
	offset = seq * sample_align;
	return shm + DSP_RING_BASE + offset;
}

int main(int argc, char **args)
{
	unsigned long seq = blocks_from;
	u8 *buf;
	u8 *p;

	parse_command_line(argc, args);

	if (dspshm_init(dsp_device, dsp_offset, dsp_size, dsp_sample)) {
		fprintf(stderr, "fatal: cannot itialize shared memory\n");
		exit(EXIT_FAILURE);
	}

	writel(0, shm + DSP_MAGIC);
	writel(0, shm + DSP_SEQNUM);
	writel(0, shm + DSP_VERSION);
	writel(dsp_size, shm + DSP_SHM_SIZE);
	writel(sample_size, shm + DSP_SAMPLE_SIZE);
	writel(samples_per_group, shm + DSP_SAMPLES_PER_GROUP);
	writel(ring_entries, shm + DSP_RING_ENTRIES);
	writel(DSP_MAGIC_VALUE, shm + DSP_MAGIC);

	p = xmalloc(dsp_sample);
	for (;;) {
		writel(seq, shm + DSP_SEQNUM);

		if (blocks_to && seq >= blocks_to)
			break;
		if (interval)
			usleep(interval);

		memset(p, seq & 0xff, dsp_sample);
		buf = dsp_get_buffer(seq);
		memcpy_toio(buf, p, dsp_sample);
		seq++;
	}
	writel(seq, shm + DSP_IGROUP);
	return 0;
}
